# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.providers.common.compat.version_compat import AIRFLOW_V_2_10_PLUS

if TYPE_CHECKING:
    from airflow.providers.standard.operators.python import (
        _SERIALIZERS,
        PythonOperator,
        ShortCircuitOperator,
        get_current_context,
    )
else:
    try:
        from airflow.providers.standard.operators.python import (
            _SERIALIZERS,
            PythonOperator,
            ShortCircuitOperator,
            get_current_context,
        )
    except ModuleNotFoundError:
        from airflow.operators.python import (
            PythonOperator,
            ShortCircuitOperator,
            get_current_context,
        )

        if AIRFLOW_V_2_10_PLUS:
            from airflow.operators.python import _SERIALIZERS


__all__ = ["PythonOperator", "_SERIALIZERS", "ShortCircuitOperator", "get_current_context"]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.common.compat.version_compat import AIRFLOW_V_2_10_PLUS, AIRFLOW_V_3_0_PLUS


def _get_asset_compat_hook_lineage_collector():
    from airflow.lineage.hook import get_hook_lineage_collector

    collector = get_hook_lineage_collector()

    if all(
        getattr(collector, asset_method_name, None)
        for asset_method_name in ("add_input_asset", "add_output_asset", "collected_assets")
    ):
        return collector

    # dataset is renamed as asset in Airflow 3.0

    from functools import wraps

    from airflow.lineage.hook import DatasetLineageInfo, HookLineage

    DatasetLineageInfo.asset = DatasetLineageInfo.dataset

    def rename_asset_kwargs_to_dataset_kwargs(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            if "asset_kwargs" in kwargs:
                kwargs["dataset_kwargs"] = kwargs.pop("asset_kwargs")

            if "asset_extra" in kwargs:
                kwargs["dataset_extra"] = kwargs.pop("asset_extra")

            return function(*args, **kwargs)

        return wrapper

    collector.create_asset = rename_asset_kwargs_to_dataset_kwargs(collector.create_dataset)
    collector.add_input_asset = rename_asset_kwargs_to_dataset_kwargs(collector.add_input_dataset)
    collector.add_output_asset = rename_asset_kwargs_to_dataset_kwargs(collector.add_output_dataset)

    def collected_assets_compat(collector) -> HookLineage:
        """Get the collected hook lineage information."""
        lineage = collector.collected_datasets
        return HookLineage(
            [
                DatasetLineageInfo(dataset=item.dataset, count=item.count, context=item.context)
                for item in lineage.inputs
            ],
            [
                DatasetLineageInfo(dataset=item.dataset, count=item.count, context=item.context)
                for item in lineage.outputs
            ],
        )

    setattr(
        collector.__class__,
        "collected_assets",
        property(lambda collector: collected_assets_compat(collector)),
    )

    return collector


def get_hook_lineage_collector():
    # Dataset has been renamed as Asset in 3.0
    if AIRFLOW_V_3_0_PLUS:
        from airflow.lineage.hook import get_hook_lineage_collector

        return get_hook_lineage_collector()

    # HookLineageCollector added in 2.10
    if AIRFLOW_V_2_10_PLUS:
        return _get_asset_compat_hook_lineage_collector()

    # For the case that airflow has not yet upgraded to 2.10 or higher,
    # but using the providers that already uses `get_hook_lineage_collector`
    class NoOpCollector:
        """
        NoOpCollector is a hook lineage collector that does nothing.

        It is used when you want to disable lineage collection.
        """

        # for providers that support asset rename
        def add_input_asset(self, *_, **__):
            pass

        def add_output_asset(self, *_, **__):
            pass

        # for providers that do not support asset rename
        def add_input_dataset(self, *_, **__):
            pass

        def add_output_dataset(self, *_, **__):
            pass

    return NoOpCollector()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Defines base entities used for providing lineage information."""

from __future__ import annotations

from typing import Any, ClassVar

import attr


@attr.s(auto_attribs=True)
class File:
    """File entity. Refers to a file."""

    template_fields: ClassVar = ("url",)

    url: str = attr.ib()
    type_hint: str | None = None


@attr.s(auto_attribs=True, kw_only=True)
class User:
    """User entity. Identifies a user."""

    email: str = attr.ib()
    first_name: str | None = None
    last_name: str | None = None

    template_fields: ClassVar = ("email", "first_name", "last_name")


@attr.s(auto_attribs=True, kw_only=True)
class Tag:
    """Tag or classification entity."""

    tag_name: str = attr.ib()

    template_fields: ClassVar = ("tag_name",)


@attr.s(auto_attribs=True, kw_only=True)
class Column:
    """Column of a Table."""

    name: str = attr.ib()
    description: str | None = None
    data_type: str = attr.ib()
    tags: list[Tag] = []

    template_fields: ClassVar = ("name", "description", "data_type", "tags")


# this is a temporary hack to satisfy mypy. Once
# https://github.com/python/mypy/issues/6136 is resolved, use
# `attr.converters.default_if_none(default=False)`


def default_if_none(arg: bool | None) -> bool:
    """Get default value when None."""
    return arg or False


@attr.s(auto_attribs=True, kw_only=True)
class Table:
    """Table entity."""

    database: str = attr.ib()
    cluster: str = attr.ib()
    name: str = attr.ib()
    tags: list[Tag] = []
    description: str | None = None
    columns: list[Column] = []
    owners: list[User] = []
    extra: dict[str, Any] = {}
    type_hint: str | None = None

    template_fields: ClassVar = (
        "database",
        "cluster",
        "name",
        "tags",
        "description",
        "columns",
        "owners",
        "extra",
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import functools
import importlib
from importlib import metadata

from packaging.version import Version

from airflow.exceptions import AirflowOptionalProviderFeatureException


def require_provider_version(provider_name: str, provider_min_version: str):
    """
    Enforce minimum version requirement for a specific provider.

    Some providers, do not explicitly require other provider packages but may offer optional features
    that depend on it. These features are generally available starting from a specific version of such
    provider. This decorator helps ensure compatibility, preventing import errors and providing clear
    logs about version requirements.

    Args:
        provider_name: Name of the provider e.g., apache-airflow-providers-openlineage
        provider_min_version: Optional minimum version requirement e.g., 1.0.1

    Raises:
        ValueError: If neither `provider_name` nor `provider_min_version` is provided.
        ValueError: If full provider name (e.g., apache-airflow-providers-openlineage) is not provided.
        TypeError: If the decorator is used without parentheses (e.g., `@require_provider_version`).
    """
    err_msg = (
        "`require_provider_version` decorator must be used with two arguments: "
        "'provider_name' and 'provider_min_version', "
        'e.g., @require_provider_version(provider_name="apache-airflow-providers-openlineage", '
        'provider_min_version="1.0.0")'
    )
    # Detect if decorator is mistakenly used without arguments
    if callable(provider_name) and not provider_min_version:
        raise TypeError(err_msg)

    # Ensure both arguments are provided and not empty
    if not provider_name or not provider_min_version:
        raise ValueError(err_msg)

    # Ensure full provider name is passed
    if not provider_name.startswith("apache-airflow-providers-"):
        raise ValueError(
            f"Full `provider_name` must be provided starting with 'apache-airflow-providers-', "
            f"got `{provider_name}`."
        )

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                provider_version: str = metadata.version(provider_name)
            except metadata.PackageNotFoundError:
                try:
                    # Try dynamically importing the provider module based on the provider name
                    import_provider_name = provider_name.replace("apache-airflow-providers-", "").replace(
                        "-", "."
                    )
                    provider_module = importlib.import_module(f"airflow.providers.{import_provider_name}")

                    provider_version = getattr(provider_module, "__version__")

                except (ImportError, AttributeError, ModuleNotFoundError):
                    raise AirflowOptionalProviderFeatureException(
                        f"Provider `{provider_name}` not found or has no version, "
                        f"skipping function `{func.__name__}` execution"
                    )

            if provider_version and Version(provider_version) < Version(provider_min_version):
                raise AirflowOptionalProviderFeatureException(
                    f"Provider's `{provider_name}` version `{provider_version}` is lower than required "
                    f"`{provider_min_version}`, skipping function `{func.__name__}` execution"
                )

            return func(*args, **kwargs)

        return wrapper

    return decorator
